/*
   Copyright (C) 1997-2001 Id Software, Inc.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 */

#include "tvm_local.h"

#include "tvm_chase.h"
#include "tvm_clip.h"
#include "tvm_client.h"

//====================
//TVM_CanChase
//====================
static qboolean TVM_CanChase( edict_t *ent, edict_t *target )
{
	assert( ent && ent->local && ent->r.client );

	if( !target->r.inuse || target->local || !target->r.client )
		return qfalse;

	if( PLAYERNUM( target ) == target->relay->playernum )
		return qfalse;

	return qtrue;
}

//====================
//TVM_ChaseChange
// Can be called when no chase target set, will then find from beginning or end (depending on the step)
//====================
static void TVM_ChaseChange( edict_t *ent, int step )
{
	int i;
	edict_t	*target, *start;

	assert( ent && ent->local && ent->r.client );
	assert( step != 0 );

	if( !ent->relay->maxclients )
		return;

	if( ent->r.client->chase.target )
	{
		start = ent->r.client->chase.target;
	}
	else
	{
		if( step > 0 )
		{
			start = ent->relay->edicts + ent->relay->maxclients; // so we find first
		}
		else
		{
			start = ent->relay->edicts + 1; // so we find last
		}
	}

	i = ENTNUM( start );

	do
	{
		i += step;
		if( i < 1 )
		{
			i = ent->relay->maxclients;
		}
		else if( i > ent->relay->maxclients )
		{
			i = 1;
		}

		target = ent->relay->edicts + i;
		if( TVM_CanChase( ent, target ) )
		{
			ent->r.client->chase.target = target;
			break;
		}
	}
	while( target != start );
}

//====================
//TVM_Cmd_ChaseNext
//====================
void TVM_Cmd_ChaseNext( edict_t *ent )
{
	assert( ent && ent->local && ent->r.client );

	if( !ent->r.client->chase.active )
		return;

	TVM_ChaseChange( ent, 1 );
}

//====================
//TVM_Cmd_ChasePrev
//====================
void TVM_Cmd_ChasePrev( edict_t *ent )
{
	assert( ent && ent->local && ent->r.client );

	if( !ent->r.client->chase.active )
		return;

	TVM_ChaseChange( ent, -1 );
}

//====================
//TVM_ChaseCamMode
//====================
static void TVM_ChaseCamMode( edict_t *ent )
{
	gclient_t *client;

	assert( ent && ent->local && ent->r.client );

	client = ent->r.client;

	// Is already in chasecam
	if( client->chase.active )
		return;

	// locate a chase target
	if( !client->chase.target )
	{
		TVM_ChaseChange( ent, 1 );
		if( !client->chase.target )
		{
			//failed: stay as observer
			TVM_CenterPrintMsg( ent->relay, ent, "No one to chase" );
			return;
		}
	}

	client->chase.active = qtrue;
}

//====================
//TVM_SpectatorMode
//====================
static void TVM_SpectatorMode( edict_t *ent )
{
	assert( ent && ent->local && ent->r.client );

	// was in chasecam
	if( ent->r.client->chase.active )
		ent->r.client->chase.active = qfalse;

	ent->r.client->ps.pmove.pm_type = PM_SPECTATOR;
	ent->r.client->ps.pmove.pm_flags &= ~PMF_NO_PREDICTION;
}

//====================
//TVM_Cmd_ChaseCam
//====================
void TVM_Cmd_ChaseCam( edict_t *ent )
{
	assert( ent && ent->local && ent->r.client );

	TVM_ChaseCamMode( ent );
}

//====================
//TVM_Cmd_SwitchChaseCamMode - Used by cgame for switching mode when clicking the mouse button
//====================
void TVM_Cmd_SwitchChaseCamMode( edict_t *ent )
{
	assert( ent && ent->local && ent->r.client );

	if( ent->r.client->chase.active )
	{
		TVM_SpectatorMode( ent );
	}
	else
	{
		TVM_ChaseCamMode( ent );
	}
}

//====================
//TVM_ChaseTargetRemoved
//====================
void TVM_ChaseTargetRemoved( edict_t *ent )
{
	edict_t *target;

	assert( ent && ent->local && ent->r.client );

	if( !ent->r.client->chase.active )
	{
		ent->r.client->chase.target = NULL;
		return;
	}

	target = ent->r.client->chase.target;
	TVM_ChaseChange( ent, 1 );
	if( ent->r.client->chase.target == target )
	{
		TVM_CenterPrintMsg( ent->relay, ent, "No one to chase" );
		TVM_SpectatorMode( ent );
		ent->r.client->chase.target = NULL;
	}
}

//=================
//TVM_ChaseClientEndSnapFrame
//=================
void TVM_ChaseClientEndSnapFrame( edict_t *ent )
{
	edict_t *target, *spec;

	assert( ent && ent->local && ent->r.client );
	assert( ent->r.client->chase.active );

	target = ent->r.client->chase.target;

	// cam controls
	if( ent->r.client->chase.keyNext == qtrue )
	{                                           // change player
		ent->r.client->chase.keyNext = qfalse;
		TVM_ChaseChange( ent, 1 );
	}

	spec = ent->relay->edicts + ent->relay->playernum + 1;

	// copy target playerState to me
	ent->r.client->ps = target->r.client->ps;

	// chasecam uses PM_CHASECAM
	ent->r.client->ps.pmove.pm_type = PM_CHASECAM;
	ent->r.client->ps.pmove.pm_flags |= PMF_NO_PREDICTION;
	ent->r.client->ps.POVnum = ENTNUM( target );
	if( TVM_ClientIsZoom( ent ) )
	{
		ent->r.client->ps.fov = ent->r.client->pers.zoomfov;
	}
	else
	{
		ent->r.client->ps.fov = ent->r.client->pers.fov;
	}

	VectorCopy( target->s.origin, ent->s.origin );
	VectorCopy( target->s.angles, ent->s.angles );

	GClip_LinkEntity( ent->relay, ent );
}
